;;;	SWEEPLINE
;;;
;;;	Questa  una versione AutoLisp del famoso algoritmo di calcolo dei punti di
;;;	intersezione di un gruppo di N segmenti nel piano. L'algoritmo di base  quello
;;;	di Bentley e Ottmann illustrato da Preparata e Shamos nel testo: Computational
;;;	Geometry an introduction. Questa versione  implememtata da Daniele Piazza in
;;;	AutoLisp standard della versione 13 di AutoCad.
;;;
;;;
;;;	This is an AUTOLISP version of the famous algorithm that find all intersection point
;;;	of n line segment by "SWEEPLINE" algorithm.
;;;	"Given n line segments in the plane, report all points where a pair of line segments
;;;	intersect." We assume that each line segment is represented by giving the coordinates of
;;;	its two endpoints and bulge (for future...).
;;;	For test my program you can draw a group of lines (intersect and not...) in AUTOCAD screen
;;;	and type "SLTEST". Then you must select the line segments, the program draw a virtual
;;;	point for each intersection point.
;;;   You can use this software for your functions.


;;;	Revisione: XX-XX-XX
;;;	Inserita una funzione di rotazione degli oggetti in ingresso in modo da poter
;;;	analizzare anche linee verticali. La rotazione avviene con un angolo "random".
;;;
;;;
;;;	Programmato da:	Daniele Piazza
;;;	per:			Mechanical Solution S.r.l.
;;;				Soluzioni software per la meccanica
;;;				Via Catalani, 3 Parma -ITALY-
;;;				Tel. +39 0521 872638
;;;				email: mecsol@mec-sol.com
;;;
;;;
;;;	Thank you and remeber:
;;;
;;;	"Ci sono pi cose in cielo e in terra Orazio, di quante se ne sognano nella la tua filosofia."
;;;
;;;								Shakespeare   -   Hamlet [Atto I Scena V]
;;;
;;;	Goodbye !!!





;;; Verifica se due valori numerici hanno lo stesso segno


(defun SAME_SIGN (a b)
 (or (and (minusp a)(minusp b))
     (and (not (minusp a))(not (minusp b)))
 )
)



;;;	Elimina l'oggetto ELE dalla lista LST

(defun KILL (ele lst)
 (apply 'append (subst nil (list ele) (mapcar 'list lst)))
)




;;;	Estrae il codice dxf CODE da un'entit NAME

(defun DXFN (code name)
  (cdr (assoc code (entget name)))
)


(defun ROTVEC (vect angolo / vx vy a11 a21 nvx nvy)
 (setq vx (car vect)
       vy (cadr vect)
       a11 (cos angolo)
       a21 (sin angolo)
       nvx (+ (* a11 vx)(* (- a21) vy))
       nvy (+ (* a21 vx)(* a11 vy))
 )
 (list nvx nvy (caddr vect))
)



;;;	ROTOBJ
;;;
;;;	Questa funzione ruota di un certo angolo un oggetto, se l'angolo  zero non esegue il
;;;	calcolo e restituisce l'oggetto cos come .

(defun ROTOBJ (OBJ ALPHA)
 (if (not (zerop alpha))
  (list (rotvec (car obj) alpha)
        (rotvec (cadr obj) alpha)
        (caddr obj)
  )
  obj
 )
)


(defun RUOTALISTA (LISTA ALPHA)
 (mapcar '(lambda (obj)(rotobj obj alpha)) lst)
)





;;;	TWO-LINES...che calcola i punti di intersezione di due linee


(defun TWO-LINES (OBJ1 OBJ2 /  x1 x2 x3 x4 y1 y2 y3 y4 x y fuzzy
			       a1 a2 b1 b2 denom p1 p2 p3 p4
			       res r1 r2 r3 r4 c1 c2 num
		   )
 (setq fuzzy 1e-8
       p1 (car obj1)  p3 (car obj2)
       p2 (cadr obj1) p4 (cadr obj2)
 )
 (if (or (equal p1 p3 fuzzy)(equal p1 p4 fuzzy)
	 (equal p2 p3 fuzzy)(equal p2 p4 fuzzy)
     )
  (setq res nil)
  (progn
   (setq x1 (car p1)  x2 (car p2)  x3 (car p3)  x4 (car p4)
	 y1 (cadr p1) y2 (cadr p2) y3 (cadr p3) y4 (cadr p4)

	 a1 (- y2 y1)
	 b1 (- x1 x2)
	 c1 (- (* x2 y1) (* x1 y2))
	 r3 (+ (* a1 x3) (* b1 y3) c1)
	 r4 (+ (* a1 x4) (* b1 y4) c1)
   )
   (if (and (not (equal r3 0.0 fuzzy))
	    (not (equal r4 0.0 fuzzy))
	    (same_sign r3 r4)
       )
    (setq res nil)                        ;no intersezione
    (progn
     (setq a2 (- y4 y3)
	   b2 (- x3 x4)
	   c2 (- (* x4 y3)(* x3 y4))
	   r1 (+ (* a2 x1) (* b2 y1) c2)
	   r2 (+ (* a2 x2) (* b2 y2) c2)
     )
     (if (and (not (equal r1 0.0 fuzzy))
	      (not (equal r2 0.0 fuzzy))
	      (same_sign r1 r2)
	 )
      (setq res nil)                      ;no intersezione
      (progn
       (setq denom (- (* a1 b2) (* a2 b1)))
       (if (equal denom 0.0 fuzzy)
	(setq res nil)                    ;collinear
	(setq num (- (* b1 c2) (* b2 c1))
	      x  (/ num denom)
	      num (- (* a2 c1) (* a1 c2))
	      y  (/ num denom)
	      res (list x y (caddr p1))
	)
       );i
      );p
     );i
    );p
   );i
  );p
 );i
; (print res)
 res
)


;;;	LSTSORT
;;;
;;;	Riordina una lista analizzando il primo elemento di ogni oggetto della lista

(defun LSTSORT (LST / LST1 a b LO HI OK fuzzy)
 (setq fuzzy 1e-8)
 (repeat  (/ (length LST) 2)
  (setq LO (car LST)
        HI nil
        LST1 nil
  )
  (foreach item (cdr LST)
   (and (< (car item) (car LO))
        (setq OK LO
              LO item
              item OK
        )
   )
   (and (> (car item) (car HI))
        (setq OK HI
              HI item
              item OK
        )
   )
   (setq LST1 (cons item LST1))
  );f
  (setq A (cons LO A)
        B (cons HI B)
        LST (cdr (reverse LST1))
  )
 );r
 (append (reverse A) LST B)
)





;;;	X_SORT

;;;	Funzione di ordinamento secondo l'algoritmo shellshort
;;;	modificato per soddisfare il nostro ordinamento. Ordina una lista
;;;	di punti secondo x crescente (ordinamento primario) e y crescente
;;;	(ordinamento secondario).


(defun X_SORT (LST / LST1 a b LO HI OK fuzzy)
 (setq fuzzy 1e-8)
 (repeat  (/ (length LST) 2)
  (setq LO (car LST)
        HI nil
        LST1 nil
  )
  (foreach pt (cdr LST)
   (and (or (< (car pt) (car LO))
            (and (equal (car pt) (car LO) fuzzy)
                 (< (cadr pt) (cadr LO))
            )
        )
        (setq OK LO
              LO pt
              pt OK
        )
   )
   (and (or (> (car pt) (car HI))
            (and (equal (car pt) (car HI) fuzzy)
                 (> (cadr pt) (cadr HI))
            )
        )
        (setq OK HI
              HI pt
              pt OK
        )
   )
   (setq LST1 (cons pt LST1))
  );f
  (setq A (cons LO A)
        B (cons HI B)
        LST (cdr (reverse LST1))
  )
 );r
 (append (reverse A) LST B)
)


;;;	OBJXSORT
;;;
;;;	Questa funzione riordina i singoli oggetti in modo tale che il primo punto stia a sinistra
;;;	(in basso) rispetto al secondo.


(defun OBJXSORT (lista / fuzzy lout)
 (setq fuzzy 1e-8)
 (foreach obj lista
  (if (or (< (caar obj) (caadr obj))
          (and (equal (caar obj) (caadr obj) fuzzy)
               (< (cadar obj)(cadadr obj))
          )
      )
   (setq lout (cons obj lout))
   (setq lout (cons (list (cadr obj) (car obj) (- (caddr obj))) lout))
  )
 )
 lout
)


;;;	FINDOBJ
;;;
;;;	Questa funzione si preoccupa di cercare nel profilo un oggetto che abbia il punto
;;;	iniziale (se il valore pos  0) o il punto finale (se il valore pos  1) uguali al
;;;	punto passato in ingresso.



(defun FINDOBJ (PT LST POS / fuzzy obj flag)
 (setq fuzzy 1e-8
       flag nil
 )
 (while (and (setq obj (car lst))
             (not (equal pt (nth pos obj) fuzzy))     ;0 = primo punto, 1 = secondo punto
        )
  (setq lst (cdr lst))
 )
 obj
)


;;;	X-CROSS
;;;
;;;	Interseca un segmento con una linea verticale e restituisce la y di intersezione



(defun X-CROSS (x obj / y fuzzy)
 (setq fuzzy 1e-8)
 (if (equal (caar obj) (caadr obj) fuzzy)
  (cadar obj)
  (+ (cadar obj)(/ (* (- x (caar obj))(- (cadadr obj)(cadar obj)))
                   (- (caadr obj)(caar obj))
                )
  )
 )
)


;;;	REORDER-L
;;;
;;;	Questa funzione riordina la lista di sweep line status al valore di x
;;;	che le viene passato in ingresso.


(defun REORDER-L (x L)
 (mapcar 'cadr (lstsort
   (mapcar '(lambda (item1 item2)
             (list item1 item2)
            )
    (mapcar '(lambda (obj)
                        (x-cross x obj)
                       )
                L
               )
    L
   )
  )
 )
)


;;;	ABOVE / BELOW
;;;
;;;	Queste due funzioni restituiscono rispettivamente il segmento superiore e quello
;;;	inferiore nella lista L, rispetto a quello passato in ingresso.

(defun ABOVE (s L)
 (car (cdr (member s L)))
)

(defun BELOW (s L)
 (car (cdr (member s (reverse L))))
)


;;;	SWAP
;;;
;;;	Questa funzione restituisce la lista L con i due oggetti s1 e s2 scambiati di posto



(defun SWAP (s1 s2 L)
 (setq L (subst "TMP" s1 L)
       L (subst s1 s2 L)
       L (subst s2 "TMP" L)
 )
)

;;;	NODE_IN_STACK
;;;
;;;	Questa funzione verifica se un punto appartiene ad una lista oppure no. Equivalente
;;;	a MEMBER, ma concede una tolleranza numerica variabile


(defun NODE_IN_STACK (pt lst / fuzzy)
 (setq fuzzy 1e-8)
 (apply 'or (mapcar '(lambda (a) (equal pt a fuzzy)) lst))
)


;;;	SWEEPLINE
;;;
;;;	Per il momento ci limitiamo ad implementare il caso di segmenti di retta
;;;	riservandoci di allargare il codice a segmenti di arco (monotoni) in un secondo
;;;	tempo, quando il codice sar numericamente consolidato.


(defun SWEEPLINE (LISTA / pqueue pt L s A tmp px lout s1 s2 s3 s4 B)
 (setq lista (objxsort lista)
       pqueue (x_sort (append (mapcar 'car lista)(mapcar 'cadr lista)))
 )
 (while (setq pt (car pqueue))
  (setq pqueue (cdr pqueue))
  (cond

   ((setq s (findobj pt lista 0))		;left endpoint
    (setq L (reorder-l (car pt) (cons s L))
          s1 (above s L)		;above
          s2 (below s L)             ;below
    )
    (if (and s1 (setq tmp (two-lines s s1)))
     (setq A (cons (list tmp s1 s) A))
    )
    (if (and s2 (setq tmp (two-lines s s2)))
     (setq A (cons (list tmp s s2) A))
    )
   );c1

   ((setq s (findobj pt lista 1))		;right endpoint
    (setq L (reorder-l (car pt) L)
          s1 (above s L)
          s2 (below s L)
    )
    (if (and s1 s2
             (setq tmp (two-lines s1 s2))
             (> (car tmp) (car pt))
        )
     (setq A (cons (list tmp s1 s2) A))
    )
    (setq L (kill s L))
   );c2

   (T
    (setq tmp (assoc pt B)
          s1 (cadr tmp)
          s2 (caddr tmp)
          s3 (above s1 L)
          s4 (below s2 L)
    )
    (if (and s3 (setq tmp (two-lines s3 s2)))
     (setq A (cons (list tmp s3 s2) A))
    )
    (if (and s4 (setq tmp (two-lines s1 s4)))
     (setq A (cons (list tmp s1 s4) A))
    )
    (setq L (swap s1 s2 L))
   );end intersection


  );c
  (setq B (append B A))		Per collegare le intersezioni ai segmenti
  (while (setq tmp (car A))
   (setq A (cdr A)
         px (car tmp)
         s1 (cadr tmp)
         s2 (caddr tmp)
   )
   (if (and (not (member px pqueue))			;node_in_stack
            (> (car px) (car pt))
       )
    (setq lout (cons (list px s1 s2) lout)
          pqueue (x_sort (cons px pqueue))			;non ottimizzato
    )
   )
  )
 )
 lout
)

(defun GET_TIME ()
  (* 86400 (getvar "tdusrtimer"))
)

;;;	RANDNUM
;;;
;;; Random number generation function - based on the linear
;;; congruential method as presented in Doug Cooper's book
;;; Condensed Pascal, pp. 116-117.
;;; Returns a random number between 0 and 1.

(defun RANDNUM()
 (if (not seed)(setq seed (getvar "DATE")))
 (setq seed (rem (+ (* 25173 seed) 13849) 65536))
 (* (/ seed 65536) 100)
)


(defun FUZZY_ROTATE (lista / ang)
 (setq ang (randnum)
       lista (ruotalista lista ang)
 )
 (list lista ang)
)




;       QUESTA FUNZIONE DISEGNA UN PUNTO SULLO SCHERMO TRAMITE VETTORI. LA MISURA DEL PUNTO E'
;       PROPORZIONATA ALLA VISTA CORRENTE


(defun DRAW_PT (pt color / dd p1 p2 p3 p4)
  (setq dd (/ (getvar "VIEWSIZE") 100.0)
	p1 (polar pt 0.785398 dd)
	p2 (polar pt 2.356194 dd)
	p3 (polar pt -2.356194 dd)
	p4 (polar pt -0.785398 dd)
  )
  (grdraw p1 p3 color)
  (grdraw p2 p4 color)
 )



;;; MAIN FUNCTION
;;;
;;;	In yy var you can find all intersection points. In xx var you can find intersection
;;;	points with their segments.

(defun C:SLTEST (/ ss lst ang)
 (setq ss (ssget))
 (while (setq ent (ssname ss 0))
  (setq ss (ssdel ent ss)
        lst (cons (list (dxfn 10 ent)
                        (dxfn 11 ent)
                        0.0
                  )
                  lst
            )
  )
 )
 (setq t1 (get_time))
 (setq ang (randnum)
       lst (ruotalista lst ang)
 )
 (setq yy (mapcar 'car (setq xx (sweepline lst))))    ;;xx deve essere ruotata
 (setq yy (mapcar '(lambda (pt)
                    (rotvec pt (- ang))
                   )
           yy
 )        )
 (setq t2 (get_time))
 (setq total_time (- t2 t1))
 (foreach pt yy
  (draw_pt pt 2)
 )
 (princ (strcat "\nComputing time: " (rtos total_time 2 2) " sec"))
 (princ)
)
(prompt "\nType SLTEST to start...")


;;;--------------------------------------------------------------------
;;; Comments by Reini Urban, 24.09.97

;;; This is quite academic code and should be used, if you
;;; expect a huge number of lines to intersect and you are looking for all
;;; intersection points. (sweepline) should run in O(n*logn), but the shell
;;; sort in AutoLISP is too slow. but with vlisp it should work significantly
;;; faster.

;;; Otherwise a greedy method like below is sufficient.
;;; O(n*2),  exactly: O(n*(n-1))
;;; test each line with all other lines
;;; segments: ((p1 p2)(p1 p2)...)
(defun greedy-all-intersections (segments / seg1 rest seg2 pt pts)
  (setq rest segments)
  (foreach seg1 segments
    (setq rest (cdr rest))
    (foreach seg2 rest
      (if (setq pt (inters (car seg1)(cadr seg1)(car seg2)(cadr seg2) T))
        (setq pts (cons pt pts)))))
  pts
)

;;; Extensions:
;;; It is possible to add bulged segments too by extending the data structure
;;; "segment" with (p1 p2 bulge) for curved segments and
;;; use a line-arc resp. arc-arc intersection method for those.
;;; Then a straight segment is still representable with (p1 p2),
;;; but you will have to take care that the new methods will return a list of
;;; intersection points instead of just one point or nil.

;;; line-arc and arc-arc intersection methods are e.g. in GLNADS.C of the
;;; AutoCAD SDK 2.0 and may be easily ported to AutoLISP.
;;; Another method is with Vill 3.0 and R14 where
;;; it is possible to use the VB method (vla-intersect obj1 obj2) which works
;;; with splines too.